/*
	unterra.monstoro.MotionTracker library -- by Gughunter


	Key steps to using the library:

	-- Override obj/ummt_blip/Init and call SetGraphics to define the icon(s) that
		will represent blips.

	-- Decide how you'll call ummt_ReportMove when things move. turf/Entered or
		atom/movable/Move are good candidates.

	-- When you create trackers, be sure to call tracker.SetOwner and tracker.SetView.
		SetView uses the same scheme as world.view, so for a standard BYOND world,
		tracker.SetView(5) will make blips appear over the actual objects -- probably
		not too useful!

	-- When a mob loses a tracker, call tracker.SetOwner(null) to clear the mob's screen.


	Cautions:

	-- This library assumes world.view and client.view are a single number -- i.e., the
		view is square. Fancy settings like "11x9" are not supported!

	-- This library uses screen objects, which have a reputation for hogging bandwidth.
		Do adequate testing to ensure that your game will not lag.

*/


obj/ummt_blip
	icon_state = "ummt_noblip"
	layer = MOB_LAYER + 0.5
	var/blip_state = "blip"
	var/atom/owner
	var/ummt_tracker/tracker
	var/flick_mode = 1


	proc/Init()
		// Override this to change blip's initial appearance. The "owner" and "tracker"
		// vars may be useful to help you determine context.


	proc/Update()
		/* Override this to adjust the blip's appearance over time. Note that Init and Update
		might be useful for completely non-graphical purposes, such as adding a musical tone
		that represents the closest proximity to the player.
		*/


	proc/SetFlickMode(newValue)
		flick_mode = newValue


	proc/SetGraphics(myIcon, myBlipState, myLayer)
		if(myIcon) icon = myIcon
		if(myBlipState) blip_state = myBlipState
		if(myLayer) layer = myLayer


	New(atom/A, ummt_tracker/T)
		. = ..()
		owner = A
		tracker = T
		Init()


ummt_virtualLoc
	var
		atom/base
		xOffset
		yOffset


	proc/SetPosition(atom/A, newXOffset, newYOffset)
		base = A
		xOffset = newXOffset
		yOffset = newYOffset


	proc/GetCenterX()
		return base.x + xOffset


	proc/GetCenterY()
		return base.y + yOffset


ummt_tracker
	var/client/owner
	var/list/blips = list()
	var/ummt_virtualLoc/eye
	var/view = 5
	var
		minScreenX; minScreenY; maxScreenX; maxScreenY


	proc/SetOwner(mob/M)
		if(M == owner) return

		if(owner && !M)
			Clear()
			owner = null
			return

		if(!istype(M) || !M.client) return

		owner = M.client
		SetVisibleScreen(1, 1, 2 * owner.view + 1, 2 * owner.view + 1)


	proc/SetEye(atom/A, xOffset, yOffset)
		eye.SetPosition(A, xOffset, yOffset)


	proc/SetView(newValue)
		view = newValue


	proc/SetVisibleScreen(minX, minY, maxX, maxY)
		minScreenX = minX
		minScreenY = minY
		maxScreenX = maxX
		maxScreenY = maxY


	proc/Clear()
		for(var/B in blips)
			del blips[B]
			blips -= B


	proc/ValidZLevel(blipLevel)
		//Default is to register only blips on the same Z level.
		//You can override this to handle the blip's Z level however you like.
		//Example: For blips to register if they're within one floor, the expression
		//would be (abs(eye.base.z - blipLevel) <= 1).
		if(eye.base.z == blipLevel) return 1


	proc/yxcabax(a, b, c, x, y)
		return (y - x) * (c - a) / (b - a) + x


	proc/RegisterMove(atom/movable/A)
		if(!owner) return
		if(!ValidZLevel(A.z)) return

		var/obj/ummt_blip/B = blips[A]
		if(!B)
			B = new(A, src)
			blips[A] = B
			owner.screen += B
		B.Update()

		var/newX = yxcabax(eye.GetCenterX() - view, eye.GetCenterX() + view, A.x, \
			1, 2 * owner.view + 1)
		var/newY = yxcabax(eye.GetCenterY() - view, eye.GetCenterY() + view, A.y, \
			1, 2 * owner.view + 1)
		var/xoff = round((newX - round(newX)) * 32)
		var/yoff = round((newY - round(newY)) * 32)

		if((newX >= minScreenX) && (newX <= maxScreenX) && \
		(newY >= minScreenY) && (newY <= maxScreenY))
			newX = round(newX)
			newY = round(newY)
			B.screen_loc = "[newX]:[xoff],[newY]:[yoff]"

			if(B.flick_mode) flick(B.blip_state, B)
			else B.icon_state = B.blip_state

		else B.icon_state = "ummt_noblip"


	New()
		eye = new()
		var/ummt_handler/H = ummt_GetHandler()
		H.AddTracker(src)


ummt_handler
	var/list/trackers = list()


	proc/RegisterMove(atom/movable/A)
		for(var/ummt_tracker/T in trackers)
			T.RegisterMove(A)


	proc/AddTracker(ummt_tracker/T)
		if(istype(T)) trackers += T


proc/ummt_ReportMove(atom/movable/A)
	var/ummt_handler/H = ummt_GetHandler()
	H.RegisterMove(A)


proc/ummt_GetHandler()
	var/global/ummt_handler/H
	if(!H) H = new()
	return H
